@@ -5,6 +5,8 @@
 Download Amazon EBS snapshots.
 */
 
+use crate::block_device::get_block_device_size;
+use async_trait::async_trait;
 use futures::stream::{self, StreamExt};
 use indicatif::ProgressBar;
 use rusoto_ebs::{Ebs, EbsClient};
@@ -14,10 +16,11 @@ use snafu::{ensure, OptionExt, ResultExt, Snafu};
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::io::SeekFrom;
+use std::os::unix::fs::FileTypeExt;
 use std::path::{Path, PathBuf};
 use std::sync::{Arc, Mutex};
 use tempfile::NamedTempFile;
-use tokio::fs::OpenOptions;
+use tokio::fs::{self, OpenOptions};
 use tokio::io::{AsyncSeekExt, AsyncWriteExt};
 
 #[derive(Debug, Snafu)]
@@ -57,37 +60,36 @@ impl SnapshotDownloader {
         progress_bar: Option<ProgressBar>,
     ) -> Result<()> {
         let path = path.as_ref();
-        let _ = path.file_name().context(error::ValidateFileName { path })?;
-        let target_dir = path
-            .parent()
-            .context(error::ValidateParentDirectory { path })?;
+        let _ = path
+            .file_name()
+            .context(error::ValidateFileNameSnafu { path })?;
 
         // Find the overall volume size, the block size, and the metadata we need for each block:
         // the index, which lets us calculate the offset into the volume; and the token, which we
         // need to retrieve it.
-        let Snapshot {
-            volume_size,
-            block_size,
-            blocks,
-        } = self.list_snapshot_blocks(snapshot_id).await?;
+        let snapshot: Snapshot = self.list_snapshot_blocks(snapshot_id).await?;
 
-        // Create a temporary file and extend it to the required size.
-        let temp_file = NamedTempFile::new_in(target_dir)
-            .context(error::CreateTempFile { path: target_dir })?;
-        let temp_file_len = volume_size * GIBIBYTE;
-        let temp_file_len = u64::try_from(temp_file_len).with_context(|| error::ConvertNumber {
-            what: "temp file length",
-            number: temp_file_len.to_string(),
-            target: "u64",
-        })?;
-        temp_file
-            .as_file()
-            .set_len(temp_file_len)
-            .context(error::ExtendTempFile {
-                path: temp_file.as_ref(),
-            })?;
-        let temp_path = temp_file.into_temp_path();
+        let mut target = if BlockDeviceTarget::is_valid(path).await? {
+            BlockDeviceTarget::new_target(path)?
+        } else {
+            // If not block assume file for now
+            FileTarget::new_target(path)?
+        };
 
+        target.grow(snapshot.volume_size * GIBIBYTE).await?;
+        self.write_snapshot_blocks(snapshot, target.write_path()?, progress_bar)
+            .await?;
+        target.finalize()?;
+
+        Ok(())
+    }
+
+    async fn write_snapshot_blocks(
+        &self,
+        snapshot: Snapshot,
+        write_path: &Path,
+        progress_bar: Option<ProgressBar>,
+    ) -> Result<()> {
         // Collect errors encountered while downloading blocks, since we can't
         // return a result directly through `for_each_concurrent`.
         let block_errors = Arc::new(Mutex::new(BTreeMap::new()));
@@ -95,12 +97,13 @@ impl SnapshotDownloader {
         // We may have a progress bar to update.
         let progress_bar = match progress_bar {
             Some(pb) => {
-                let pb_length = blocks.len();
-                let pb_length = u64::try_from(pb_length).with_context(|| error::ConvertNumber {
-                    what: "progress bar length",
-                    number: pb_length.to_string(),
-                    target: "u64",
-                })?;
+                let pb_length = snapshot.blocks.len();
+                let pb_length =
+                    u64::try_from(pb_length).with_context(|_| error::ConvertNumberSnafu {
+                        what: "progress bar length",
+                        number: pb_length.to_string(),
+                        target: "u64",
+                    })?;
                 pb.set_length(pb_length);
                 Arc::new(Some(pb))
             }
@@ -109,13 +112,13 @@ impl SnapshotDownloader {
 
         // Create a context for each block that can be moved to another thread.
         let mut block_contexts = Vec::new();
-        for SnapshotBlock { index, token } in blocks {
+        for SnapshotBlock { index, token } in snapshot.blocks {
             block_contexts.push(BlockContext {
-                path: temp_path.to_path_buf(),
+                path: write_path.to_path_buf(),
                 block_index: index,
                 block_token: token,
-                block_size,
-                snapshot_id: snapshot_id.to_string(),
+                block_size: snapshot.block_size,
+                snapshot_id: snapshot.snapshot_id.clone(),
                 block_errors: Arc::clone(&block_errors),
                 progress_bar: Arc::clone(&progress_bar),
                 ebs_client: self.ebs_client.clone(),
@@ -158,18 +161,14 @@ impl SnapshotDownloader {
                 .iter()
                 .map(|(_, e)| format!("\n{}", e))
                 .collect();
-            error::GetSnapshotBlocks {
+            error::GetSnapshotBlocksSnafu {
                 error_count: block_errors_count,
-                snapshot_id: snapshot_id.to_string(),
+                snapshot_id: snapshot.snapshot_id,
                 error_report,
             }
             .fail()?;
         }
 
-        temp_path
-            .persist(&path)
-            .context(error::PersistTempFile { path })?;
-
         Ok(())
     }
 
@@ -193,23 +192,23 @@ impl SnapshotDownloader {
                 .ebs_client
                 .list_snapshot_blocks(request)
                 .await
-                .context(error::ListSnapshotBlocks { snapshot_id })?;
+                .context(error::ListSnapshotBlocksSnafu { snapshot_id })?;
 
             volume_size = response
                 .volume_size
-                .context(error::FindVolumeSize { snapshot_id })?;
+                .context(error::FindVolumeSizeSnafu { snapshot_id })?;
 
             block_size = response
                 .block_size
-                .context(error::FindBlockSize { snapshot_id })?;
+                .context(error::FindBlockSizeSnafu { snapshot_id })?;
 
-            for block in response.blocks.unwrap_or_else(Vec::new).iter() {
+            for block in response.blocks.unwrap_or_default().iter() {
                 let index = block
                     .block_index
-                    .context(error::FindBlockIndex { snapshot_id })?;
+                    .context(error::FindBlockIndexSnafu { snapshot_id })?;
 
                 let token = String::from(block.block_token.as_ref().context(
-                    error::FindBlockProperty {
+                    error::FindBlockPropertySnafu {
                         snapshot_id,
                         block_index: index,
                         property: "token",
@@ -226,6 +225,7 @@ impl SnapshotDownloader {
         }
 
         Ok(Snapshot {
+            snapshot_id: snapshot_id.to_string(),
             volume_size,
             block_size,
             blocks,
@@ -249,32 +249,35 @@ impl SnapshotDownloader {
             .ebs_client
             .get_snapshot_block(block_request)
             .await
-            .context(error::GetSnapshotBlock {
+            .context(error::GetSnapshotBlockSnafu {
                 snapshot_id,
                 block_index,
             })?;
 
-        let expected_hash = response.checksum.context(error::FindBlockProperty {
+        let expected_hash = response.checksum.context(error::FindBlockPropertySnafu {
             snapshot_id,
             block_index,
             property: "checksum",
         })?;
 
-        let checksum_algorithm = response
-            .checksum_algorithm
-            .context(error::FindBlockProperty {
+        let checksum_algorithm =
+            response
+                .checksum_algorithm
+                .context(error::FindBlockPropertySnafu {
+                    snapshot_id,
+                    block_index,
+                    property: "checksum algorithm",
+                })?;
+
+        let data_length = response
+            .data_length
+            .context(error::FindBlockPropertySnafu {
                 snapshot_id,
                 block_index,
-                property: "checksum algorithm",
+                property: "data length",
             })?;
 
-        let data_length = response.data_length.context(error::FindBlockProperty {
-            snapshot_id,
-            block_index,
-            property: "data length",
-        })?;
-
-        let block_data = response.block_data.context(error::FindBlockProperty {
+        let block_data = response.block_data.context(error::FindBlockPropertySnafu {
             snapshot_id,
             block_index,
             property: "data",
@@ -282,7 +285,7 @@ impl SnapshotDownloader {
 
         ensure!(
             checksum_algorithm == SHA256_ALGORITHM,
-            error::UnexpectedBlockChecksumAlgorithm {
+            error::UnexpectedBlockChecksumAlgorithmSnafu {
                 snapshot_id,
                 block_index,
                 checksum_algorithm,
@@ -291,7 +294,7 @@ impl SnapshotDownloader {
 
         let block_data_length = block_data.len();
         let block_data_length =
-            i64::try_from(block_data_length).with_context(|| error::ConvertNumber {
+            i64::try_from(block_data_length).with_context(|_| error::ConvertNumberSnafu {
                 what: "block data length",
                 number: block_data_length.to_string(),
                 target: "i64",
@@ -299,7 +302,7 @@ impl SnapshotDownloader {
 
         ensure!(
             data_length > 0 && data_length <= block_size && data_length == block_data_length,
-            error::UnexpectedBlockDataLength {
+            error::UnexpectedBlockDataLengthSnafu {
                 snapshot_id,
                 block_index,
                 data_length,
@@ -313,7 +316,7 @@ impl SnapshotDownloader {
 
         ensure!(
             block_hash == expected_hash,
-            error::BadBlockChecksum {
+            error::BadBlockChecksumSnafu {
                 snapshot_id,
                 block_index,
                 block_hash,
@@ -335,10 +338,10 @@ impl SnapshotDownloader {
             .write(true)
             .open(path)
             .await
-            .context(error::OpenFile { path })?;
+            .context(error::OpenFileSnafu { path })?;
 
         let offset = context.block_index * block_size;
-        let offset = u64::try_from(offset).with_context(|| error::ConvertNumber {
+        let offset = u64::try_from(offset).with_context(|_| error::ConvertNumberSnafu {
             what: "file offset",
             number: offset.to_string(),
             target: "u64",
@@ -346,9 +349,9 @@ impl SnapshotDownloader {
 
         f.seek(SeekFrom::Start(offset))
             .await
-            .context(error::SeekFileOffset { path, offset })?;
+            .context(error::SeekFileOffsetSnafu { path, offset })?;
 
-        let count = usize::try_from(data_length).with_context(|| error::ConvertNumber {
+        let count = usize::try_from(data_length).with_context(|_| error::ConvertNumberSnafu {
             what: "byte count",
             number: data_length.to_string(),
             target: "usize",
@@ -356,9 +359,9 @@ impl SnapshotDownloader {
 
         f.write_all(&block_data)
             .await
-            .context(error::WriteFileBytes { path, count })?;
+            .context(error::WriteFileBytesSnafu { path, count })?;
 
-        f.flush().await.context(error::FlushFile { path })?;
+        f.flush().await.context(error::FlushFileSnafu { path })?;
 
         if let Some(ref progress_bar) = *context.progress_bar {
             progress_bar.inc(1);
@@ -370,6 +373,7 @@ impl SnapshotDownloader {
 
 /// Stores the metadata about the snapshot contents.
 struct Snapshot {
+    snapshot_id: String,
     volume_size: i64,
     block_size: i64,
     blocks: Vec<SnapshotBlock>,
@@ -393,14 +397,179 @@ struct BlockContext {
     ebs_client: EbsClient,
 }
 
+/// Shared interface for write targets.
+#[async_trait]
+trait SnapshotWriteTarget {
+    // grow the target to the desired length
+    async fn grow(&mut self, length: i64) -> Result<()>;
+
+    // returns the file path to which blocks must be written
+    fn write_path(&self) -> Result<&Path>;
+
+    // persist the contents to disk
+    fn finalize(&mut self) -> Result<()>;
+}
+
+/// Implements file operations for block devices.
+struct BlockDeviceTarget {
+    path: PathBuf,
+}
+
+impl BlockDeviceTarget {
+    fn new_target<P: AsRef<Path>>(path: P) -> Result<Box<dyn SnapshotWriteTarget>> {
+        let path = path.as_ref();
+        Ok(Box::new(BlockDeviceTarget { path: path.into() }))
+    }
+
+    async fn is_valid<P: AsRef<Path>>(path: P) -> Result<bool> {
+        let path = path.as_ref();
+        if !path.exists() {
+            return Ok(false);
+        }
+
+        let file_meta = fs::metadata(path)
+            .await
+            .context(error::ReadFileMetadataSnafu { path })?;
+
+        if file_meta.file_type().is_block_device() {
+            Ok(true)
+        } else {
+            Ok(false)
+        }
+    }
+}
+
+#[async_trait]
+impl SnapshotWriteTarget for BlockDeviceTarget {
+    // ensures existing size >= length, but otherwise leaves untouched
+    async fn grow(&mut self, length: i64) -> Result<()> {
+        let path = self.path.as_path();
+        let block_device_size =
+            get_block_device_size(path).context(error::GetBlockDeviceSizeSnafu)?;
+
+        // Make sure the block device is big enough to hold the snapshot
+        ensure!(
+            block_device_size >= length,
+            error::BlockDeviceTooSmallSnafu {
+                block_device_size: block_device_size / GIBIBYTE,
+                needed: length / GIBIBYTE,
+            }
+        );
+
+        Ok(())
+    }
+
+    // returns the file path to which blocks must be written
+    fn write_path(&self) -> Result<&Path> {
+        Ok(self.path.as_path())
+    }
+
+    // no-op
+    fn finalize(&mut self) -> Result<()> {
+        Ok(())
+    }
+}
+
+/// Implements file operations for filesystem files.
+struct FileTarget {
+    path: PathBuf,
+    temp_file: Option<NamedTempFile>,
+}
+
+impl FileTarget {
+    fn new_target<P: AsRef<Path>>(path: P) -> Result<Box<dyn SnapshotWriteTarget>> {
+        let path = path.as_ref();
+        Ok(Box::new(FileTarget {
+            path: path.into(),
+            temp_file: None,
+        }))
+    }
+}
+
+#[async_trait]
+impl SnapshotWriteTarget for FileTarget {
+    // truncate file to desired size
+    async fn grow(&mut self, length: i64) -> Result<()> {
+        let path = self.path.as_path();
+
+        // Create a temporary file and extend it to the required size.
+        let target_dir = path
+            .parent()
+            .context(error::ValidateParentDirectorySnafu { path })?;
+
+        let temp_file = NamedTempFile::new_in(target_dir)
+            .context(error::CreateTempFileSnafu { path: target_dir })?;
+
+        let temp_file_len = length;
+        let temp_file_len =
+            u64::try_from(temp_file_len).with_context(|_| error::ConvertNumberSnafu {
+                what: "temp file length",
+                number: temp_file_len.to_string(),
+                target: "u64",
+            })?;
+
+        temp_file
+            .as_file()
+            .set_len(temp_file_len)
+            .context(error::ExtendTempFileSnafu {
+                path: temp_file.as_ref(),
+            })?;
+
+        self.temp_file.replace(temp_file);
+
+        Ok(())
+    }
+
+    fn write_path(&self) -> Result<&Path> {
+        let write_path = self
+            .temp_file
+            .as_ref()
+            .context(error::MissingTempFileSnafu {})?;
+
+        Ok(write_path.as_ref())
+    }
+
+    // persist file to destination
+    fn finalize(&mut self) -> Result<()> {
+        let temp_file = self
+            .temp_file
+            .take()
+            .context(error::MissingTempFileSnafu {})?;
+
+        let path = self.path.as_path();
+        temp_file
+            .into_temp_path()
+            .persist(path)
+            .context(error::PersistTempFileSnafu { path })?;
+
+        Ok(())
+    }
+}
+
 /// Potential errors while downloading a snapshot and writing to a local file.
 mod error {
     use snafu::Snafu;
     use std::path::PathBuf;
 
     #[derive(Debug, Snafu)]
-    #[snafu(visibility = "pub(super)")]
+    #[snafu(visibility(pub(super)))]
     pub(super) enum Error {
+        #[snafu(display("Failed to read metadata for '{}': {}", path.display(), source))]
+        ReadFileMetadata {
+            path: PathBuf,
+            source: std::io::Error,
+        },
+
+        #[snafu(display("{}", source))]
+        GetBlockDeviceSize { source: crate::block_device::Error },
+
+        #[snafu(display(
+            "Block device too small: block device size {} GiB, needed at least {} GiB",
+            block_device_size,
+            needed
+        ))]
+        BlockDeviceTooSmall { block_device_size: i64, needed: i64 },
+
         #[snafu(display("Failed to validate file name '{}'", path.display()))]
         ValidateFileName { path: PathBuf },
 
@@ -425,6 +594,9 @@ mod error {
             source: tempfile::PathPersistError,
         },
 
+        #[snafu(display("Missing temporary file"))]
+        MissingTempFile {},
+
         #[snafu(display("Failed to list snapshot blocks for '{}': {}", snapshot_id, source))]
         ListSnapshotBlocks {
             snapshot_id: String,
